/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


#include "global.h"
#include "actuators.h"
#include "helpers.h"
#include "adc.h"

// TODO: try to take out some includes:
#include "fuelcontrol.h"
#include "fuelcalc.h"
#include "comm.h"
#include "iac.h"
#include "eventqueue.h"
#include "dispatcher.h"
#include "injconf.h"
#include "ignconf.h"
#include "edis.h"
#include "benchmark.h"
#include "timing.h"

#include <inttypes.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/interrupt.h>

#include "lbus.c"
#include "injector.c"

volatile extern struct engine_t engine;




/***************************************************************************/
/*** coolant fan                                                         ***/
/***************************************************************************/
/*!
 * \ingroup g_outputs
 *
 * \brief Control for Coolant Fan Output.
 *
 * details ...
 */

void coolant_fan(void) {
	if (engine.coolant > config.fan_temp) {

#ifdef GENBOARDv3
		cli();
		digitalout(config.fan_channel, DHIGH);
		sei();
#else
#ifndef FRIED_FET_BUT_RACE_HACK
		PORTB |= _BV(CLT_FAN); // start fan
#endif
#endif

	} else if (engine.coolant <= (config.fan_temp - config.fan_hyst)) {

#ifdef GENBOARDv3
		cli();
		digitalout(config.fan_channel, DLOW);
		sei();
#else
#ifndef FRIED_FET_BUT_RACE_HACK
		PORTB &= ~_BV(CLT_FAN); // stop fan
#endif
#endif
	}
}

#ifdef GENBOARDv3
/***************************************************************************/
/*** water pump                                                          ***/
/***************************************************************************/
/*!
 * \ingroup g_outputs
 *
 * \brief Control for Water Pump Output.
 *
 * details ...
 */
void water_pump(void)
{
	if (engine.coolant > config.water_pump_temp) {
		cli();
		digitalout(config.water_pump_channel, DHIGH);
		sei();
	} else if (engine.coolant <= (config.water_pump_temp - config.water_pump_hyst)) {
		cli();
		digitalout(config.water_pump_channel, DLOW);
		sei();
	}
}
#endif


#ifdef WOT_OUTPUT_TPS_THRESHOLD
/*!
 * \ingroup g_outputs
 *
 * \brief Control for Wide Open Throttle Output.
 *
 * Call from userspace;
 * Note that stimout uses the same channel (don't call when stimout enabled)
 * TODO: add hysteresis like rpm_output()
 */
void wot_output(void)
{
	uint8_t polarity;
	if((engine.tps > WOT_OUTPUT_TPS_THRESHOLD) && (engine.rpm > config.act_wot_rpm)){
		polarity = DHIGH;
	} else {
		polarity = 0;
	}

	cli();
#ifdef GENBOARDv3	
	digitalout(config.act_wot_channel, polarity);
#else
#warning "do not enable FAST_IDLE output in config"
	// is the inversion confusing? Or do we say DHIGH is for after the NPN ?
	if(DHIGH)
		PORTB &= ~_BV(IDLEVALVE); // reused idlevalve output (off)
	else
		PORTB |= _BV(IDLEVALVE); // reused idlevalve output (on)	
#endif	// not GENBOARDv3
	
	sei();
}	// wot_output()
#endif	// WOT_OUTPUT_TPS_THRESHOLD

/*!
 * \ingroup g_outputs
 * \brief Control for RPM based Output.
 * Used for variable intake actuation.
 * Call from userspace;
 * 200 RPM histeresys by default, override with ACT_RPM_HYSTERESYS [100 RPM] unit
 */
#ifndef ACT_RPM_HYSTERESYS
#define ACT_RPM_HYSTERESYS (2)
#endif
void rpm_output(void)
{
	uint8_t polarity;
	uint8_t engine_rpm = engine.rpm;
	uint8_t threshold = config.act_rpm_rpm;
	if(engine_rpm > threshold){
		polarity = DHIGH;
	} else if(engine_rpm <= (threshold - ACT_RPM_HYSTERESYS)){
		polarity = 0;
	} else {
		return;
	}
	
	cli();
	digitalout(config.act_rpm_channel, polarity);
	sei();
}	// rpm_output()

#ifdef MISC1OUT

/*
 * Only call from userspace
 * TODO: add some hysteresis?
 */
void misc1_activity()
{
	uint8_t polarity;
	uint8_t engine_rpm = engine.rpm;
	uint8_t engine_tps = engine.tps;
	uint8_t engine_kpa = engine_kpa8();
	if (engine_rpm >= config.misc1out_minrpm &&
		engine_rpm < config.misc1out_maxrpm &&
		engine_tps >= config.misc1out_mintps &&
		engine_tps < config.misc1out_maxtps &&
		engine_kpa >= config.misc1out_minmap &&
		engine_kpa < config.misc1out_maxmap) {
		polarity=DHIGH;
	} else {
		polarity=DLOW;
	}
	cli();
	digitalout(config.misc1out_channel, polarity);
	sei();
}
#endif

#ifdef MISC2OUT
/*
 * Only call from userspace
 * TODO: add some hysteresis?
 */
void misc2_activity()
{
	uint8_t polarity;
	uint8_t engine_rpm = engine.rpm;
	uint8_t engine_tps = engine.tps;
	uint8_t engine_kpa = engine_kpa8();
	if (engine_rpm >= config.misc2out_minrpm &&
		engine_rpm < config.misc2out_maxrpm &&
		engine_tps >= config.misc2out_mintps &&
		engine_tps < config.misc2out_maxtps &&
		engine_kpa >= config.misc2out_minmap &&
		engine_kpa < config.misc2out_maxmap) {
		polarity=DHIGH;
	} else {
		polarity=DLOW;
	}
	cli();
	digitalout(config.misc2out_channel, polarity);
	sei();
}
#endif
