/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


#ifndef ADC_H
#define ADC_H

extern uint8_t adc_channel;

typedef uint16_t sample_t;

// sensor channels. Common to genboard v2 and v3:
#define VBATT  0  //Battery Voltage ADC Raw Reading - counts see config.batt_cal
#define EGO_1 1  //Exhaust Gas Oxygen ADC Raw Reading (or WBO2 sensor1)
#define TPS   2  //Throttle Position Sensor ADC Raw Reading - counts, represents 0 - 5 volts. see config.tps_low, tps_high and engine.tps
#define CLT   3  //Coolant Temperature ADC Raw Reading - counts (0 - 255)
#define MAT   4  //Manifold Air Temp ADC Raw Reading - counts (0 - 255)
#define MAP   5  //Manifold Absolute Pressure ADC Raw Reading - KPa (0 - 255)

#ifdef GENBOARDv3

#define ADC_CHANNELS 8
#define KNOCK  6  //Knock
#define EGO_2  7 // WBO2 sensor 2

#else	// not GENBOARDv3

#define ADC_CHANNELS 6

#endif	// GENBOARDv3

#define BARO   ADC_CHANNELS  //Barometer ADC Raw Reading - KPa (0 - 255) "VIRTUAL SENSOR"
#define ADC_CHANNELS_AND_BARO (ADC_CHANNELS+1)
extern sample_t sensor_filt[ADC_CHANNELS_AND_BARO];

// 8 bit compatibility getter and setter:
#define sensors8(x)	(sensor_filt[x]>>3)
#define set_sensors8(x, val)	(sensor_filt[x] = (uint16_t)val<<3)
#define SENSORS_BYTES (8*sizeof(sample_t))
// the precise getter:
#define sensors11(x) (sensor_filt[x])

// div-128 prescaler (16MHz => 8 usec clock):
#define ADCSRA_PRESCALER (_BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0))
// clearly high for WBO2 but why not set anyway ??
#define ADCSRA_ADFR (_BV(ADFR))
// AVCC connected to AREF pin internally (decoupling cap externally):
#define ADMUX_AREF (_BV(REFS0))

void startADC(void);
sample_t read_adc(uint8_t channel);
void init_adc(void);
void init_sensors(void);
void set_adc_channel(uint8_t t);
uint8_t read_adc_channel(uint8_t t);

#endif
