
/*********************************************************\
*  http://www.vems-group.org/				  *
*	BOOSTCONTROL                                      *
*  (C) 2004 [Marcell Gal, cell at x-dsl.hu                *
**********************************************************/
#include "global.h"

#ifdef BOOSTCONTROL

#include "boostcontrol.h"

#include "eventqueue.h"
#include "helpers.h"
#include "adc.h"
#include "pid.h"

#include "lbus.h"
#include "softpwm.c"
#include "injector.c"

struct pid_t boostcontrol_pid;
// struct softpwm_t boost_pwm_soft;

/* actuates the boostcontrol solenoid;
   called from mainloop.
*/
void boostcontrol_act()
{
	uint16_t boost, command16;
	uint8_t output;		// DLOW or DHIGH
	uint8_t schedule_needed;	// true if we need to schedule (the other edge)

	command16 = 0;	// just to suppress a gcc warning
	// command16 _could_ be used for schedule_needed (before += now is applied)

	/* 
		input is currently fixed:
		ADC7 EGO_2
			or 
		MAP for MAPtarget

		TODO: make it selectable, including any MCP3208 input
	*/
	if(config.boost_conf & _BV(BOOSTTARGET)){
		// MAPtarget. Good luck, Mr. Gorski!
		boost = sensors11(MAP);
	} else {
		// boosttarget: don't forget to hook up your sensor and nernst2 1x"amplifier"
		boost = sensors11(EGO_2);
	}

	// going to 12 bits:
	boost += boost;
	
	if( (boost >> 4) > (uint16_t)config.boost_minpressure ){
		uint16_t pid_result;
		// TODO: boostcut (above configurable boost) in fuelcalc.c

		pid_set_target( &boostcontrol_pid, boost_target() );
		// with pid15() high-boost will decrease command (more DLOW)
		// which is consistent with DLOW above
		pid_result = pid15( &boostcontrol_pid, boost );		
		// pid_result useful range is -2048 .. +2047 (consider this in PID config)

		// offset going to 0..4095 ( 0.5 fixed offset should be OK )
		pid_result += 2048;

		// prevent underflow, overflow.
		// 3 outputs possible: always low, always high
		// and something between
		if( pid_result >= 0x8000 || pid_result < BOOSTACT_MINTIME){
			// always OFF
			schedule_needed = 0;
			output = DLOW;
		} else if( pid_result >= 4096 ){ // compare against (4096 - BOOSTACT_MINTIME) ?
			// fully ON! good luck Mr. Gorsky!
			schedule_needed = 0;
			output = DHIGH;
		} else {
			schedule_needed = 1;	// true !!!
			output = DHIGH;
		// when pid_result is 4095, we want appr.
		// (BOOSTCONTROL_PWM_PERIOD_8MS * 2048) * 4 usec ON-time
			command16 = mult16_8( pid_result, BOOSTCONTROL_PWM_PERIOD_8MS);
			command16 >>= 1;		
		}
		
	} else {
	/*
		Don't activate boost valve below/equal a configurable boost:
		there is no need to wear the valve at cruising
		at low boost where the wastegate is known not to open anyway.
		Without this, the pid15() would raise command, so DHIGH would be likely
	*/
		schedule_needed = 0;
		output = DLOW;			
	}

	if( schedule_needed ) {
		// prepare to schedule event to reset the signal
		uint16_t now;
		now = whats_the_time();

		// command16 meaning changes at this point:
		command16 += now;
		cli();
		// engine.boost_stopat must be changed before digitalout() is called
		// otherwise a racing event could reset output right after
		// we activate the output
		engine.boost_stopat = command16;		
		sei();
	}	

	cli();
	// for inverted output set bit7 of config.boost_channel:
	digitalout( config.boost_channel, output );
	sei();

	if( schedule_needed ) {
		cli();
		event_insert( command16, EVENT_BOOSTACT );
		// without sei_nop_cli(), pending events get priority
		// before other interrupts. 
		// sei_nop_cli();
		alarm2core();
		sei();
	}
};

void boostcontrol_init()
{
	pid_init(&boostcontrol_pid, (struct pid_parameters_t*) &config.boost_pid_kp);
}
#endif	// BOOSTCONTROL
