/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


/*
TODO:
The letters of the non-standard comm. commands are a bit too random, 
fix it before it is too late!
*/


/*
  This file deserves a little explanation!

  Instead of creating a huge circular buffer for uart data, 
  a smaller circular buffer 'command_queue' are used instead. 
  This buffer contains the address and arguments to functions which 
  can fetch bytes from any place in the memory.

  eg. MegaTune sends an 'A' to request realtime variables:
  1) Uart interrupt calls comm()
  2) comm pushes the address of sendRTvar and the number of 
     bytes to send (22) onto the command queue and enables tx-interrupt.
  3) tx-interrupt reads the command queue, executes the function and
     decrements the number of bytes to send (and stores it in the 
     command queue again). If number equals zero remove the function 
     from the queue and start over. If the queue is empty, then disable 
     interrupt.
*/

   /*!
   * \addtogroup g_serial_comm Serial Communication Routines
   */
/*@{*/


#include "global.h"
#include "comm.h"
#include "ringbuf.h"
#include "storage.h"
#include "actuators.h"
#include "phasesignal.h"
#include "ve.h"
#include "iac.h"
#include "adc.h"
#include "lcd.h"
#include "ego.h"
#include "fuelcalc.h"
#include "helpers.h"
#include "benchmark.h"
#include "wbo2.h"
#include "boostcontrol.h"
#include "sipr.h"
#include "spi.h"
#include "firmware_revision.h"


#include <avr/io.h>
#include <avr/crc16.h>
#include <inttypes.h>
#include <avr/signal.h>
#include <avr/sfr_defs.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "varstr.h"

extern volatile struct engine_t engine;
// extern volatile struct time_t rtc;

extern uint8_t code_ver;

#include "arithm.c"

struct cmd_queue_t cmd_queue[CMD_QUEUE_SIZE];
uint8_t cmd_queue_tail, cmd_queue_head;

uint8_t comm_state;

char vems_string[] PROGMEM = "VEMS Genboard v3, ";


// build MegaTune compatibility table, now the variables in config_t can be
// rearranged to a reasonable order

#if (VE_SIZE_RPM == 8 && VE_SIZE_KPA == 8)
PROGMEM uint8_t *mt_config_var[] = {
  (uint8_t *)&config.cwl, 
  (uint8_t *)&config.cwh,
  (uint8_t *)&config.awev,
  (uint8_t *)&config.awc,
  (uint8_t *)&config.warmup_clt[0],
  (uint8_t *)&config.warmup_clt[1],
  (uint8_t *)&config.warmup_clt[2],
  (uint8_t *)&config.warmup_clt[3],
  (uint8_t *)&config.warmup_clt[4],
  (uint8_t *)&config.warmup_clt[5],
  (uint8_t *)&config.warmup_clt[6],
  (uint8_t *)&config.warmup_clt[7],
  (uint8_t *)&config.warmup_clt[8],
  (uint8_t *)&config.warmup_clt[9],
  (uint8_t *)&config.tpsaq[0],
  (uint8_t *)&config.tpsaq[1],
  (uint8_t *)&config.tpsaq[2],
  (uint8_t *)&config.tpsaq[3],
  (uint8_t *)&config.tpsacold,
  (uint8_t *)&config.tps_thresh,
  (uint8_t *)&config.tpsasync,
  (uint8_t *)&config.tpsdq,
  (uint8_t *)&config.ego_coolant,
  (uint8_t *)&config.ego_lag,
  (uint8_t *)&config.ego_delta,
  (uint8_t *)&config.ego_lean_limit,
  (uint8_t *)&config.req_fuel,
  (uint8_t *)&config.divider,
  (uint8_t *)&config.alternate,
  (uint8_t *)&config.injopen,
  (uint8_t *)&config.injocfuel,
  (uint8_t *)&config.injpwm,
  (uint8_t *)&config.injpwmt,
  (uint8_t *)&config.battfac,
  (uint8_t *)&config.rpmk[0],
  (uint8_t *)&config.rpmk[1],
  (uint8_t *)&config.config11,
  (uint8_t *)&config.config12,
  (uint8_t *)&config.config13,
  (uint8_t *)&config.primep,
  (uint8_t *)&config.ego_minrpm,
  (uint8_t *)&config.fastidle,
  (uint8_t *)&config.ego_target,
  (uint8_t *)&config.acmult,
  (uint8_t *)&config.mt_unused,
  (uint8_t *)&config.mt_unused,
  (uint8_t *)&config.mt_unused,
  (uint8_t *)&config.mt_unused };
#endif


/* it could be a boolean returning macro like (x & 0xFE) == 24 */
#define EGODELTA_INDEX (24)
#define EGOLEANLIMIT_INDEX (25)
#define INJOPEN (29)
#define BATTFACTOR (33)
#define INJOCFUEL (30)

#if ((defined WBO2) || (defined KNOCK_8101))
uint8_t comm_TX_plainbytes[comm_TX_plain_SIZE];
#endif


/***************************************************************************/
/*** comm                                                                ***/
/***************************************************************************/
void comm(uint8_t rx_data) {
  static uint8_t rx_offset = 0;
  static uint8_t rx_cmd = 0;

  switch (comm_state) {
  case OFFSET:
    rx_offset = rx_data;  // the offset for storing the new byte
    comm_state = VALUE;
    break;

  case VALUE:
    switch (rx_cmd) {
#ifdef MSNS
      case 'J':
	MSNSstoreConfigVar(rx_offset, rx_data);
	break;
#endif
      case 'p':
	change_page(rx_data);	// change to page 'rx_data'
	break;
      case 'n':
	storeVE(rx_offset, rx_data);
	break;
      case 'N':
	storeVEref(rx_offset, rx_data);
	break;
      case 'W':
#if (VE_SIZE_RPM != 8 || VE_SIZE_KPA != 8)
  	#warning "Table-size above 8x8 breaks megatune compat 'W' command, use paged writing 'w' instead"
#else
  	MTstoreConfigVar(rx_offset, rx_data); 
#endif
	break;
      case 'w':
	store_page(rx_offset, rx_data);
	break;
    default:
      break;
    }
    comm_state = CMD; // back to normal mode
    break;

  default:
    switch (rx_data) {
    case 'A': // send all realtime values
	pushfunc(&MTsendRTvar, MT_RTVAR_COUNT);
      break;

    case 'B': // store to eeprom
      save_variables();
      break;

    case 'C': // test communications
      pushfunc(&MTtestComm, 1);
      break;

    case 'F': // flush transmit buffer
      // NOT STANDARD MEGASQUIRT COMMAND!
      UCSR0B &= ~_BV(UDRIE0);             // disable TX interrupts
      cmd_queue_tail = cmd_queue_head = 0;
      break;
#ifdef MSNS
    case 'J': // receive new ignition parameter at 'J'+<offset>+<newbyte>
      comm_state = OFFSET;
      rx_cmd = 'J';
      break;
    case 'I': // send ignition parameters
	pushfunc(&MSNSsendConfigVar, 0x5F);
      break;
#endif
    case 'L': // send VE table
      // NOT STANDARD MEGASQUIRT COMMAND!
      pushfunc(&sendVE, VE_SIZE_RPM*VE_SIZE_KPA);  // ve.table is 16 bit
      break;

    case 'l': // send fractions from VE table
      // NOT STANDARD MEGASQUIRT COMMAND!
      pushfunc(&sendVEfraction, VE_SIZE_RPM*VE_SIZE_KPA);  // ve.table is 16 bit
      break;

    case 'm': // send VEref table
      // NOT STANDARD MEGASQUIRT COMMAND!
      pushfunc(&sendVEref, VE_SIZE_RPM*VE_SIZE_KPA);
      break;      

    case 'N': // write VE table
      // NOT STANDARD MEGASQUIRT COMMAND!
      comm_state = OFFSET;
      rx_cmd = 'N';
      break;      

    case 'n': // write VE reference table
      // NOT STANDARD MEGASQUIRT COMMAND!
      comm_state = OFFSET;
      rx_cmd = 'n';
      break;      

    case 'p': // request page number 'p'+<page>
      comm_state = VALUE;
      rx_cmd = 'p';
      break;

    case 'Q': // send code version
      pushfunc(&MTcodeVer, 1);
      break;

    case 'S': // send textual signature
      pushfunc(&sendSignature, 24);
      break;

    case 'T': // send entire config structure, and ONLY config structure (not ve tables)
      // NOT STANDARD MEGASQUIRT COMMAND!
      pushfunc(&sendConfigVar, sizeof(struct config_t));
      break;

    case 'V': // send VE table and constants
	    // 48 for 128 bytes
	    // 45 for 125 bytes
#if (VE_SIZE_RPM != 8 || VE_SIZE_KPA != 8)
	#warning "Table-size above 8x8 breaks megatune compat 'V' command, use paged writing 'v' instead"
#else
	pushfunc(&MTsendConfigVar, 45 + VE_SIZE + VE_SIZE_RPM + VE_SIZE_KPA);
#endif
      break;

    case 'W': // receive new VE or constant at 'W'+<offset>+<newbyte>
      comm_state = OFFSET;
      rx_cmd = 'W';
      break;

    case 'v': // send page
	pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);  // don't know the length yet
      break;

    case 'w': // receive value at 'w'+<offset>+<newbyte> in page debug.megatune_page
      comm_state = OFFSET;
      rx_cmd = 'w';
      break;

    default:
      break;
    }
  }
}


/***************************************************************************/
/***************************************************************************/
/*** MegaTune compatibility functions                                    ***/
/***************************************************************************/
/***************************************************************************/

uint8_t sendSignature(uint8_t i) {
  if (i < 18) { // TODO: check sizeof ?
        return pgm_read_byte( (uint8_t *)&vems_string + i );
  }
  i -= 18;
  if (i < 5)
        return *( (uint8_t *)&FIRMWARE_VERSION_STR + i );
  return 0;
}

/***************************************************************************/
/*** codeVer                                                             ***/
/***************************************************************************/
// return codeversion
uint8_t MTcodeVer(uint8_t i) {
  return BG_EMULATION;
}


/***************************************************************************/
/*** testComm                                                            ***/
/***************************************************************************/
// return low part of second-count
uint8_t MTtestComm(uint8_t i) {
  return (rtc.sec & 0xFF);
}



/***************************************************************************/
/*** getMCPdataHIGH                                                      ***/
/***************************************************************************/
// does not start a adc conversion, just masks out the high byte of latest conversion 
// and returns if the channel data is not 0xffff

// channels range from 0-7 internally (from 1 to 8 according to datasheet)

uint8_t getMCPdataHIGH(uint8_t i) {

	uint16_t t;
	uint8_t r = 0;

	if(i < 8) {
		t = mcp3208_data[i];
		if (t != 0xffff) 
	//		r = (uint8_t) t >> 8;
			r = t >> 8;
	}
	
	return r;	
}


/***************************************************************************/
/*** getMCPdataLOW                                                       ***/
/***************************************************************************/
// starts an mpc channel adc conversion on the appropriate channel
// and if conversion is finished return low byte of adc result, else 0

// channels range from 0-7 internally (from 1 to 8 according to datasheet)

uint8_t getMCPdataLOW(uint8_t i) {

	uint16_t t;
	uint8_t r = 0;

	if(i < 8) {
		t = mcp3208_data[i];
		if (t != 0xffff) 
			r = (uint8_t) t & 0xFF;
		read_mcp3208(i);
	}
	
	return r;	
}




/***************************************************************************/
/*** sendRTvar                                                           ***/
/***************************************************************************/
// convert datastructures to megasquirt compatible format

uint8_t MTsendRTvar(uint8_t i) {
//	uint8_t a, b;

  switch (i) {
  case 0:
    return (uint8_t)rtc.sec;  //secl

  case 1:
/* TODO:
  check this.
  "squirt" is apparently a bitmap of the something:
  ....BBAAba
  where
  capital A, B : ????
  lowercase a, b: if the given channels are actually squirting
  // each squirt_t has its own status field, merge these fields
    a = inj_port1.status;
    b = inj_port2.status;

    return ((a & 0x01) 
    		| ((a & 0x06) << 1)
           	| ((b & 0x01) << 1) 
           	| ((b & 0x06) << 3));
*/

  case 2:  
// remove the 'idle' bit to avoid confusing mstweak: return (engine.status & (uint8_t)~_BV(idle));  //engine
	  return engine.status;  //engine
  case 3:  return sensors8(BARO);  //baro

  // high resolution MAP
  case 4: return engine.kpa250 >> 8;
  case 5: return engine.kpa250 & 0xFF;

  case 6: return engine.mat;   //mat
  case 7: return engine.coolant;   //clt
  case 8: return engine.tps;   //tps
  case 9: return engine.batt;  //batt
  case 10: // o2
#ifdef WBO2
	if (WBO2_SENSOR_CONFIGURED)
		return wbo2.Ip_output;
	else
#endif
		return ego_voltage();  //ego 
  case 11: return 50+((corr.ego*100)>>8);	//egocorr
  case 12: return corr.air;       //aircorr
  case 13: return corr.warm >= 155 ? 255 : corr.warm+100;      //warmcorr

  // high resolution RPM
  case 14: return engine.rpm_vhr >> 8;
  case 15: return engine.rpm_vhr & 0xFF;

  case 16: return mult16_8(inj_port[0].pwcalc, 41)>>10;   //pw
  case 17: return corr.tpsaccel;  //tpsaccel
  case 18: return corr.baro;      //barocorr
  case 19: {	// calculate from the stored high-res value (/100 and railing is needed)
		uint16_t ge;
		ge = mult16_8(corr.gammae_1k5, 17) >> 8;	// 1% error in reporting...

		if ((ge >> 8) != 0)  // check for overflow
			return 0xFF;
		else
			return (uint8_t)ge;
	}
  case 20: return corr.ve_hp>>8;

// should we use this style ? 
// #define MT_RT_OFFSET_XXX (21)
// #define MT_RT_OFFSET_IGNADV (MT_RT_OFFSET_XXX+1)

  case 21: return engine.dwell>>4;	// 64 usec resolution (possibly but unlikely overflowing)
  case 22: return engine.ignadv>>1;

  case 23: return iac.position;

// make sure 16 bit variables start on even offsets:

// raw egt data, do scaling externally (in megatune based on egtofs and egtcal)
  case 24: return getMCPdataHIGH(4); // EGT1 HIGH
  case 25: return getMCPdataLOW(4);  // EGT1 LOW
  case 26: return getMCPdataHIGH(3); // EGT2 HIGH
  case 27: return getMCPdataLOW(3);  // EGT2 LOW
  
// also raw adc data for te other sensors 
  case 28: return getMCPdataHIGH(6); // EBP HIGH
  case 29: return getMCPdataLOW(6);  // EBP LOW
  case 30: return getMCPdataHIGH(7); // FP HIGH
  case 31: return getMCPdataLOW(7);  // FP LOW
  case 32: return getMCPdataHIGH(5); // MISC1 HIGH
  case 33: return getMCPdataLOW(5);  // MISC1 LOW
  case 34: return getMCPdataHIGH(0); // MISC2 HIGH midopa
  case 35: return getMCPdataLOW(0);  // MISC2 LOW
  case 36: return getMCPdataHIGH(1); // MISC3 HIGH midopa
  case 37: return getMCPdataLOW(1);  // MISC3 LOW
  case 38: return getMCPdataHIGH(7); // MISC4 HIGH
  case 39: 
  	return getMCPdataLOW(7);  // MISC4 LOW


	// TODO: reserved status bytes like WBO2 activated and alike
  case 40: return 0;
  case 41:
	if (debug.enable_flags & _BV(7)){
		pushfunc(&MTsendRTvar, MT_RTVAR_COUNT);
	}
	return 0;
    
 default: return 0;              // 
  }
}

static inline uint8_t REINDEX(uint8_t addr)
{
#if (VE_SIZE_RPM == 8)
		return addr;
#else
		return (VE_SIZE_RPM/8)*(addr % 8) + (addr/8)*VE_SIZE_RPM; 
		// (VE_SIZE_RPM/8) *  because ...
#endif
}

// std. megasquirt protocol is only defined for a 8x8 table
#define MT_VE_SIZE_RPM 8
#define MT_VE_SIZE_KPA 8
#define MT_VE_SIZE MT_VE_SIZE_RPM * MT_VE_SIZE_KPA

#define RPMRANGE_OFFSET (MT_VE_SIZE + 36)
#define KPARANGE_OFFSET (RPMRANGE_OFFSET + 8)

#define RPMRANGE_END RPMRANGE_OFFSET + 8
#define KPARANGE_END KPARANGE_OFFSET + 8

/***************************************************************************/
/*** MTstoreConfigVar  (old, non-paging command)                         ***/
/***************************************************************************/

#if (VE_SIZE_RPM == 8 && VE_SIZE_KPA == 8)
void MTstoreConfigVar(uint8_t addr, uint8_t value) {
	uint8_t my_addr;
	
	if (addr >= 48 + MT_VE_SIZE + MT_VE_SIZE_RPM + MT_VE_SIZE_KPA)	// send a bugreport to the developers of the control program :)
		return;	// we DON'T want a loose pointer

    if (addr < VE_SIZE) {
		// write value to addressed table (usually VE reference table)
		debug.megatune_table[ addr ] = value;

		if(debug.megatune_table != &ve.ref_table[0]) // don't mess with ve if another table is addressed
			return;

		// set the fraction to 127 (evenly hard increasing/decreasing VE value)
		ve.ve_table[ addr ] = (value << 8) | 127;
		// might be a good idea to interpolate the neighbour points
		return;
    }
	
	if ((addr >= RPMRANGE_OFFSET) && (addr < (RPMRANGE_END))) {	// rpmrangeve[8]
		my_addr = addr - RPMRANGE_OFFSET;
		ve.rpm_range[my_addr] = value;
		return;
    }
	
	if ((addr >= KPARANGE_OFFSET) && (addr < (KPARANGE_END))) {	// kparangeve[8]
		my_addr = addr - KPARANGE_OFFSET;
		ve.kpa_range[my_addr] = value;
		return;
    }

	my_addr = addr - MT_VE_SIZE;
	if (my_addr >= (KPARANGE_END-VE_SIZE)) {
		my_addr -= (MT_VE_SIZE_RPM + MT_VE_SIZE_KPA);
	}

	if (my_addr == INJOPEN || my_addr == BATTFACTOR || my_addr == INJOCFUEL) { // we store 16usec resolution, not 100
		uint16_t t16;
		t16 = value*100;	// multiply by 100/16 and ignore overflow!
		value = (t16 >> 4) + 3;	// appr.: t16 + 50 so rounding correctly 
		// Note: +1 would also prevent the decreasing injopen value in megatune
	} else if(my_addr == EGODELTA_INDEX || my_addr == EGOLEANLIMIT_INDEX){
		if(value >= 50) // max limit
			value = 125;
		else 
			value = (value*41)>>4;	// * 256/100
	}
	*((uint8_t*)pgm_read_word(mt_config_var + my_addr)) = value;

}
#endif // (VE_SIZE_RPM == 8 && VE_SIZE_KPA == 8) 

#ifdef MSNS
void MSNSstoreConfigVar(uint8_t addr, uint8_t value) {
	if (addr >= 64)
		return;	// we DON'T want a loose pointer

	if (addr < 64) {
		uint8_t t;
		uint16_t adv;
		adv = (value*90) >> 6;
		if (adv>>8)
			adv = 0xff;

		t = REINDEX(addr);
		ign.adv_table[ t ] = (uint8_t)adv;
		return;
	}
}
#endif

/***************************************************************************/
/*** MTsendConfigVar (old, non-paging command)                           ***/
/***************************************************************************/

#if (VE_SIZE_RPM == 8 && VE_SIZE_KPA == 8)
uint8_t MTsendConfigVar(uint8_t addr) {
    uint8_t my_addr;
    // MegaTune compatibility, the first 64 bytes are the VE table.
    // In the 16x16 table, we want every other point in both the rpm and kpa direction
	if (addr < MT_VE_SIZE){
		if(debug.megatune_table == &ve.ref_table[0])
			return ve.ve_table[ addr ] >> 8;	// servicing from hi-res ve table
		else 
			return debug.megatune_table[addr];
	}

	if ((addr >= RPMRANGE_OFFSET) && (addr < (RPMRANGE_OFFSET + VE_SIZE_RPM)))	// rpmrangeve[8]
		return ve.rpm_range[addr - RPMRANGE_OFFSET];

	if ((addr >= KPARANGE_OFFSET) && (addr < (KPARANGE_OFFSET + VE_SIZE_KPA)))	// kparangeve[8]
		return ve.kpa_range[addr - KPARANGE_OFFSET];

    // if the size of the VE table are different from 64, then we must fix
    // some addresses (megatune compatibility)

	my_addr = addr - MT_VE_SIZE; // sizeof(VE)
	if (my_addr >= (KPARANGE_END-MT_VE_SIZE)) {
		my_addr -= (MT_VE_SIZE_RPM + MT_VE_SIZE_KPA);
	}

	if (my_addr == INJOPEN || my_addr == BATTFACTOR || my_addr == INJOCFUEL)	// adjust injopen *= 16/100
		return (*((uint8_t*)pgm_read_word(mt_config_var + my_addr)) * 41)>>8;

	if(my_addr == EGODELTA_INDEX || my_addr == EGOLEANLIMIT_INDEX){
		uint8_t t;
		t = *((uint8_t*)pgm_read_word(mt_config_var + my_addr));

		if(t < 3)	// we don't want to return 0 for egodelta (for lean_limit: who cares?)
			return 1;
		else 
			return (t*100) >> 8; // * 100/256
	}
    //    return *( (uint8_t *)&config + my_addr );
    return *((uint8_t*)pgm_read_word(mt_config_var + my_addr));
}
#endif // (VE_SIZE_RPM == 8 && VE_SIZE_KPA == 8)

#ifdef MSNS
uint8_t MSNSsendConfigVar(uint8_t addr) {
	if (addr < 64){
		uint8_t t;
		t = REINDEX(addr);
		return (ign.adv_table[t]*182) >> 8;
	}

	if ((addr >= 64) && (addr < 72))	// rpmrangeve[8]
		return ve.rpm_range[addr - 64];

	if ((addr >= 72) && (addr < 80))	// kparangeve[8]
		return ve.kpa_range[addr - 72];

	return 0;
}
#endif


/* Dump the table 
* private higher byte holds the table identifier char,
* private lower byte holds the table-line
* line higher 4 bits contain the counter (eg 0..7)
* line lower 4 bits contain the maxvalue of counter (eg 7)
* count (conservatively) holds the character inside the line
*/
uint8_t TableDump(uint8_t cnt){
	uint16_t t;
	t = (uint16_t)(cmd_queue[cmd_queue_head].private);
	char table; 
	uint8_t line, linemax, *p, len;
	uint8_t block_printlen, block_bytes, line_end;

	table = t>>8;
	line = t & 0xFF;
	linemax = line & 0x0F;
	line >>= 4;
#define HEADER_LENGTH (5)
	if(cnt < HEADER_LENGTH){ // line beginning
		switch(cnt){
			case 0: return table;
			case 1: return '[';
			case 2: return hex_lo(line);
			case 3: return ']';
			case 4: return '=';
		}	// switch
	}	// if
		
	if(cnt >= 0xF0){
		// similar to state 0, but 0 cannot be used again for some reason...
		cmd_queue[cmd_queue_head].count = 0;
		return table;	
	}
	
	block_printlen = 3;
	block_bytes = 1;

	switch(table){ // this should get from a place common with menu
		case 'j':
#ifdef VE_TABLE_FIX
			// 16 bit table:
			block_printlen = 5;
			block_bytes = 2;
			p=((uint8_t*)ve.ve_table) + 1;
#else
			p=ve.ref_table;
#endif
			len = VE_SIZE_RPM;
			break;
		case 'k':
			p=ve.kpa_range;
			len = VE_SIZE_KPA;
			break;
		case 'l':
			p=ve.lambda;
			len = VE_SIZE_RPM;
			break;
		// either ifndef NOIGN is not needed, or needed in menu.c too
		// I allow the read-write of 'n' table for other applications...
		case 'n':
			p=ign.adv_table;
			len = VE_SIZE_RPM;
			break;
#ifdef HWMAP
		case 'h':
			p=hwmap.table;
			len = HWMAP_SIZE;
			break;
#endif
#ifdef BOOSTCONTROL
		case 'b':
			p = boost.boostmaxtarget_table;
			len = VE_SIZE_RPM;
			break;
		case 't':
			p = boost.tpstransfer_table;
			len = TPSTRANSFERTABLE_SIZE;
			break;
#endif
		case 'r':
		default:
			p=ve.rpm_range;
			len = VE_SIZE_RPM;
		// TODO: dump learnt values and difference of learnt and ve_ref
	}

	// we must determine how long the line is for the given table:
	// 3 * len or 5 *len for a 16 bit table 
	// (-1 in any case, since no space before first, see the '\r' branch)
	line_end = HEADER_LENGTH + len * block_printlen;	
	if(cnt == line_end){ // shell we finish or go to next line
		if(line != linemax){ // not finished
			line++;	// could be made faster, but ....
			cmd_queue[cmd_queue_head].private = (void*)((table<<8) + (line<<4) + linemax);
			cmd_queue[cmd_queue_head].count = 0xF0-1; // will be incremented (0-1 doesnt work)
		} else { // FF would be better?
			cmd_queue[cmd_queue_head].count = 0xFE; // set to len => THE END
		}
		return '\n';
	}
	
	if(cnt == line_end - 1){
		return '\r';
	}
	
	cnt -= (HEADER_LENGTH - 1);

	p += len * line * block_bytes;

	while(cnt >= block_printlen){
		cnt -= block_printlen;
		p += block_bytes;
	}

	switch(cnt){	// every byte is printed to 3 characters
		case 0: return ' ';
		case 1: return hex_hi(*p);
		case 2: return hex_lo(*p);
		case 3: return hex_hi(*(p-1));	// only for 16 bit table (little endian)
		case 4: return hex_lo(*(p-1));			
		default: return ' ';
	}
}

/* Dump the config structure:
 * We keep track of line number in private.
 * position inside the line is in count, but somewhat tricky 
 * Note that len is 0xFE because of pushfunc() hack, although we specified 0xFF
 */
uint8_t ConfigDump(uint8_t cnt) {
	uint8_t c, i, *str; 
	uint16_t idx;
	PGM_VOID_P P;
	// not always needed, but almost:
	// if(cnt == 0xF7) cmd_queue[cmd_queue_head].count = cnt = 0; // needed because of the len-1 hack
	
	idx = (uint16_t)(cmd_queue[cmd_queue_head].private);
	P = (PGM_VOID_P)(varstr + idx);
	i = 0;
	while((str=(void*)PRG_RDW(P)) == NULL){
		P -= 2;  // addresses are two-byte
		i++;
	}
#define LAST_ELEMENT (sizeof(struct config_t)-1)
	if(cnt < 0xF8){
		c = PRG_RDB(str + cnt);
		if(c) {
			return c;
		} else { // end of string
		// we need the [.]= if i>0 or the next pointer is NULL but not at the very end
			if(i || (((void*)PRG_RDW(P+2)) == NULL && idx < LAST_ELEMENT)){ // no need for [i]=.. just =..
				cmd_queue[cmd_queue_head].count = 0xF8 - 1 ; // will be incremented!
				return '[';
			} else {
				cmd_queue[cmd_queue_head].count = 0xFB - 1 ; // will be incremented!
				return '=';
			}
		}	
	}
	switch (cnt){
		case 0xF8: return hex_lo(i);
		case 0xF9: return ']';
		case 0xFA: return '=';
		case 0xFB: return hex_hi(*((uint8_t *)(&config) + idx));
		case 0xFC: return hex_lo(*((uint8_t *)(&config) + idx));
		case 0xFD: return '\r';
		case 0xFE: // let's see: are we at the very end ?
			if((uint16_t)(++ cmd_queue[cmd_queue_head].private) <= LAST_ELEMENT ){
				// cmd_queue[cmd_queue_head].count = 0xFE;	// end of the show
				// = cmd_queue[cmd_queue_head].len 
				cmd_queue[cmd_queue_head].count = 0xFF; // show must go on next line
			};
			return '\n';
	}
	return 'z';
}
/***************************************************************************/
/***************************************************************************/
/*** End of MegaTune compatibility functions                             ***/
/***************************************************************************/
/***************************************************************************/





/***************************************************************************/
/*** sendConfigVar                                                       ***/
/***************************************************************************/
uint8_t sendConfigVar(uint8_t i) {
  if (i < sizeof(struct config_t))
    return *( (uint8_t *)&config + i );
  else 
    return 0;
}





// these functions are for the 16x16 VE tables

/***************************************************************************/
/*** sendVE                                                              ***/
/***************************************************************************/
uint8_t sendVE(uint8_t addr) {
  return ve.ve_table[addr] >> 8;
}

/***************************************************************************/
/*** sendVEfraction                                                      ***/
/***************************************************************************/
uint8_t sendVEfraction(uint8_t addr) {
  return ve.ve_table[addr];
}

/***************************************************************************/
/*** sendVEref                                                           ***/
/***************************************************************************/
uint8_t sendVEref(uint8_t addr) {
  // TODO: should this be fooled or always send ve.ref_table ?
  return debug.megatune_table[addr];
}

uint8_t srl_direct( uint8_t i ){
	switch(i){
//		case 0: return 10;
//		case 1: return 'd';
		case 0: {
			uint16_t t = (uint16_t) cmd_queue[cmd_queue_tail].private;
			return t>>8;
		}
		case 1: {
			uint16_t t = (uint16_t) cmd_queue[cmd_queue_tail].private;
			return t & 0xff;
		}
		default: return 'z';
	}
}

/***************************************************************************/
/***                           SIP-R Functions                           ***/
/***************************************************************************/

/*
* Sends out a SIP-R packet and calculates the CRC16 as it goes for the last
* part of the packet. The packet info struct is held in the .private member
*/
uint8_t srl_sipr_packet( uint8_t i){
  uint8_t c;  // char we will send
  uint8_t len=cmd_queue[cmd_queue_head].len;  // length of this packet

  switch (i) {
    case 0: // header byte return directly to avoid applying CRC
      cmd_queue[cmd_queue_head].privw=0;
      return SIPR_HEADER;
    case 1: // sequence number
      c=cmd_queue[cmd_queue_head].seq;
      break;
    case 2: // packet type always found at func(0)
      if (cmd_queue[cmd_queue_head].privatefunc!=NULL)
        c=cmd_queue[cmd_queue_head].privatefunc(0); // callout for the packet content
      else
        c='X'; // something BAD happened!
      break;
    case 3: // packet data length calculated from the length subtracting the headers and CRC
      c=len-6;
      break;
    default: // process any data and the CRC bytes
      if (i==len-2) // CRC byte 1
        return cmd_queue[cmd_queue_head].privch.a;
      else 
        if (i==len-1) // CRC byte 2
          return cmd_queue[cmd_queue_head].privch.b;
        else
          if (cmd_queue[cmd_queue_head].privatefunc!=NULL)
            c=cmd_queue[cmd_queue_head].privatefunc(i-3); // callout for the packet content
          else 
            c='X'; // something BAD happened!
  }
  // update the CRC
  cmd_queue[cmd_queue_head].privw=_crc16_update(cmd_queue[cmd_queue_head].privw, c);
  return c;
}

/**********************************************
* function for adding SIPR send functions. Different from normal enqueue commands...
* this hard codes the call to srl_sipr_packet and sets the function as the packet
* specific content generator. The last number passed is the packet sequence number
*/
uint8_t pushfuncsipr(uint8_t (*f)(uint8_t i), uint8_t len, uint8_t seq) {
  uint8_t next_tail;

  next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return -1; // whoops, queue is full

  // we always call srl_sipr_packet
  cmd_queue[cmd_queue_tail].func = &srl_sipr_packet;
  cmd_queue[cmd_queue_tail].len = len;
  cmd_queue[cmd_queue_tail].count = 0;
  cmd_queue[cmd_queue_tail].privatefunc =f;
  cmd_queue[cmd_queue_tail].seq = seq;
  cmd_queue_tail = next_tail;

//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return 0;
}

/***************************************************************************/
/***                       End of SIP-R Functions                        ***/
/***************************************************************************/


/************************************
* generic serial logging function.
* uses private from the actual command as a pointer (array base). 
* Unfortunately we need to know the string-length in advance 
* For constant(-length) strings this is OK.
* We could set a (global or static variable is simplest) flag
* to notify the caller about the end of transmission.
* This way other comm functions could be left untouched.
************************************/
uint8_t srl_generic(uint8_t addr)
{
	return (cmd_queue[cmd_queue_head].private)[addr]; 
}


/* same as above, but private points to PROGMEM (flash), eg. PSTR("...") */
uint8_t srl_generic_P(uint8_t addr)
{
	return PRG_RDB(&(cmd_queue[cmd_queue_head].private[addr])); 
}

// helper funtion used to access flash above the 16 bit address limit
// TODO: check that it works
uint8_t srl_generic_P_upper(uint8_t addr)
{
	uint32_t v;
	v = 0x10000 | (uint16_t)&cmd_queue[cmd_queue_head].private[addr];
	return pgm_read_byte_far(v);
}

/***************************************************************************/
/*** storeVE                                                             ***/
/***************************************************************************/
void storeVE(uint8_t addr, uint8_t value) {
  if (addr < (VE_SIZE_RPM*VE_SIZE_KPA)) {
    ve.ve_table[addr] = (value << 8) | 127;
  }
}

/***************************************************************************/
/*** storeVEref                                                          ***/
/***************************************************************************/
void storeVEref(uint8_t addr, uint8_t value) {
  if (addr < (VE_SIZE_RPM*VE_SIZE_KPA)) {
    debug.megatune_table[addr] = value;
  }
}


void change_page(uint8_t page) {
	if (page > 7) { // TODO: make the 7 automatic
		page = MT_PAGE_IDX_DUMMY;
	}
	debug.megatune_page = page;
}

/* ===============================================
 * make sure MT_PAGE_IDX_... and pages_size match.
 * avoid changing these since they break several tuningsoftwares
 */

#define MT_PAGE_PTR_DUMMY (uint8_t *)&config

PROGMEM uint8_t *pages_ptr[] = {
 	(uint8_t *)&config

#ifdef HWMAP
 	,(uint8_t *)&hwmap.table
#else	//  sizeof(hwmap) bytes _must_ be transmitted to megatune
 	, MT_PAGE_PTR_DUMMY
#endif

#ifdef BOOSTCONTROL
	,(uint8_t *)&boost.tpstransfer_table
#else // no BOOSTCONTROL
 	, MT_PAGE_PTR_DUMMY
#endif
 	, MT_PAGE_PTR_DUMMY
 	,(uint8_t *)&ve.ref_table
 	,(uint8_t *)&ve.lambda
 	,(uint8_t *)&ign.adv_table
	,(uint8_t *)&ve.lambda
};

PROGMEM uint8_t pages_size[] = {
 	sizeof(struct config_t)
 	, HWMAP_BYTES

#ifdef BOOSTCONTROL
	,sizeof(boost.tpstransfer_table) + sizeof(boost.boostmaxtarget_table)
#else 
	, MT_PAGE_SIZE_DUMMY
#endif

	, MT_PAGE_SIZE_DUMMY
 	,sizeof(ve.ref_table)  // ve.ve_table will be sent but this many bytes :-)
 	,sizeof(ve.lambda)
 	,sizeof(ign.adv_table)
 	,sizeof(ve.lambda)
};

#define MEGATUNE_PAGE_SIZE(page) (pgm_read_byte(pages_size + page))
#define MEGATUNE_PAGE_POINTER(page) ((uint8_t*)pgm_read_word(pages_ptr + page))

/* only makes sense for commands that send data to megatune
 * (read: not for those that store)
 */
static inline void terminate_transfer( void ) {
	cmd_queue[cmd_queue_head].len = 0;
}

/* 
 * send config and tables to megatune.
 * For page 4 the (learnt) VE table is sent,
 * not the ve.ref_table reference table
 */
 
#define MT_LAMBDA_OFFSET 73
#define LAMBDA_INT_OFFSET 200 

uint8_t send_page(uint8_t i) {
	uint8_t val, size, page;

	// a little clumsy, but avoids gcc warning
	page = (uint8_t)((uint16_t)cmd_queue[cmd_queue_head].private);
	size = MEGATUNE_PAGE_SIZE(page);

	if( i < size ){
		if( page == MT_PAGE_IDX_VE ) {
			// send upper byte of (learnt) ve.ve_table
			return ve.ve_table[i] >> 8;
		} 

		uint8_t *tbl = MEGATUNE_PAGE_POINTER(page) ;
		val = tbl[i];

		if( page == MT_PAGE_IDX_LAMBDAVIEW ) {
			uint16_t mt_val;
			mt_val = mult16_8((val + LAMBDA_INT_OFFSET), 72) >> 7;
			mt_val = (mult16_8(div_65535_x(mt_val), 72) >> 7) - MT_LAMBDA_OFFSET;
 			return mt_val & 0xFF;
		} else {

			if(page == MT_PAGE_IDX_BOOST) { 
				// after t[] there comes b[]:
				if( i > TPSTRANSFERTABLE_SIZE ){
					val = boost.boostmaxtarget_table[ i - TPSTRANSFERTABLE_SIZE ];
				}
			}

			if (( page < MT_PAGE_IDX_VE || page > MT_PAGE_IDX_LAMBDAVIEW) && ( i >= size-1 )) {
		// for the 4..6 pages we pad rpm_range and kpa_range arrays to the end
		// but all other pages end normally:
				terminate_transfer();
			}
		}

		return val;
	}
	
	// note that only pages 4..6 can get to this point

	// pad rpm_range and kpa_range arrays to the end	
	i -= size;
	if (i < VE_SIZE_RPM) {
		return ve.rpm_range[i];
	}

	i -= VE_SIZE_RPM;

	if (i < VE_SIZE_KPA ){
		val = ve.kpa_range[i];
	}
	
	if (i >= (VE_SIZE_KPA - 1)) {
		terminate_transfer();
	}

	return val;
}

/* caller is responsible for idx < VE_SIZE_KPA */
void kpa_range_store( uint8_t idx, uint8_t value ){
	ve.kpa_range[idx] = value;
}

/* caller is responsible for idx < VE_SIZE_RPM */
void rpm_range_store( uint8_t idx, uint8_t value ){
	ve.rpm_range[idx] = value;
}

/* for VE_TABLE_PAGE the 
 * learnt ve.ve_table is also written, 
 * besides ve.ref_table
 */
void store_page(uint8_t addr, uint8_t value) {
	uint8_t size, page;

	page = debug.megatune_page;
	size = MEGATUNE_PAGE_SIZE(page);

	if (addr < size) {
		uint8_t *tbl;

		if( page == MT_PAGE_IDX_VE ) {
			uint16_t t16 = (value<<8) | 0x80;
			// writing to hi-res ve.ve_table
			ve.ve_table[addr] = t16;
		} else if(page == MT_PAGE_IDX_BOOST) {
			// after t[] there comes b[]:
			if( addr > TPSTRANSFERTABLE_SIZE ){
				boost.boostmaxtarget_table[ addr - TPSTRANSFERTABLE_SIZE ] = value;
				return;
			}
		} else if(page == MT_PAGE_IDX_LAMBDAVIEW) { 
			uint16_t int_val;
			int_val = mult16_8((value + MT_LAMBDA_OFFSET), 72) >> 7; 
			int_val = mult16_8(div_65535_x(int_val) , 72) >> 7;
 			value = (int_val - LAMBDA_INT_OFFSET) & 0xFF;
 		}
 	
		// the ve.ref_table is written besides the hi-res ve.ve_table above:
		tbl = MEGATUNE_PAGE_POINTER(page);
		tbl[addr] = value;
		return;
	}

	/* 
	 * rpm_range and kpa_range arrays live at the end
	 */
	if ( page < MT_PAGE_IDX_VE || page > MT_PAGE_IDX_LAMBDAVIEW ) {
		// terminate_transfer() is meaningless (actually harmful) here !
		return;
	} 

	addr -= size;
	if ( addr < VE_SIZE_RPM ) {
		rpm_range_store( addr, value );
		return;
	}

	addr -= VE_SIZE_RPM;	

	if (addr < VE_SIZE_KPA ) {
		kpa_range_store( addr, value );
	}
}

/***************************************************************************/
/*** pushfunc                                                            ***/
/***************************************************************************/
uint8_t pushfunc(uint8_t (*f)(uint8_t i), uint8_t len) {
  uint8_t next_tail;

  next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return -1; // whoops, queue is full

  /* len starts at 0 */
//	if (len > 0) // this looks like a hack - it is! IT WAS!
//		len = len - 1;

  cmd_queue[cmd_queue_tail].func = f;
  cmd_queue[cmd_queue_tail].len = len;
  cmd_queue[cmd_queue_tail].count = 0;
  cmd_queue[cmd_queue_tail].private = 0;

  cmd_queue_tail = next_tail;
  
//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return 0;
}

/**********************************************
* same as above, but parameter can be specified 
*/
uint8_t pushfuncpar2(uint8_t (*f)(uint8_t i), uint8_t len, const uint8_t * par, uint8_t p2) {
  uint8_t next_tail;

  next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return -1; // whoops, queue is full

  cmd_queue[cmd_queue_tail].func = f;
  cmd_queue[cmd_queue_tail].len = len;
  cmd_queue[cmd_queue_tail].count = 0;
  cmd_queue[cmd_queue_tail].private = (void*)par;
  cmd_queue[cmd_queue_tail].privch.a = p2;
  cmd_queue_tail = next_tail;
  
//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return 0;
}

// compatibility function with less parameters
uint8_t pushfuncpar(uint8_t (*f)(uint8_t i), uint8_t len, const uint8_t * par) {
	return pushfuncpar2(f, len, par, 0);
}
/***************************************************************************/
/*** initUART                                                            ***/
/***************************************************************************/
void initUART(void) {

	  comm_state = CMD;
	  
	/* rx enable, tx enable */
  //  UCSR0B = _BV(RXCIE0) | _BV(UDRIE0) | _BV(RXEN0) | _BV(TXEN0);
  UCSR0B = _BV(RXCIE0) | _BV(RXEN0) | _BV(TXEN0);
  /* 8 databit, 1 stopbit */
  UCSR0C = _BV(UCSZ01) | _BV(UCSZ00);

  /* 9600 baud @16MHz */
  UBRR0H = 0;
  UBRR0L = 103;

  cmd_queue_tail = cmd_queue_head = 0;

}


/***************************************************************************/
/*** INTERRUPT: UART rx complete ***/
/***************************************************************************/
// a small receivebuffer is still waiting to be implemented
SIGNAL(SIG_UART0_RECV) {

  // check status register for frame or overrun errors
  if ( UCSR0A & (_BV(FE0) | _BV(DOR0)) ) { 
    uint8_t dummy;
    dummy = UDR0; // just read uart and throw away the result
    return;
  } else {
	ringbuf_push(&comm_recbuf, UDR0);
    // comm(UDR0);	// TODO: put to a buffer, which menu.c consumes and calls back comm() if necessary
  }
}


/***************************************************************************/
/*** INTERRUPT: UART tx empty ***/
/***************************************************************************/
SIGNAL(SIG_UART0_DATA) {
	uint16_t d;

	d = ringbuf_shift(&comm_sndbuf);

	if (d>>8) { // buffer empty
		UCSR0B &= ~_BV(UDRIE0);    // disable TX interrupts
	} else {
		UDR0 = d;
	}

/*
  //call function and get data to send
  UDR0 = (*cmd_queue[cmd_queue_head].func)(cmd_queue[cmd_queue_head].count);

  if (cmd_queue[cmd_queue_head].count < cmd_queue[cmd_queue_head].len) {
    cmd_queue[cmd_queue_head].count++;
  } else {
    // we're done with this function

    cmd_queue_head = (cmd_queue_head + 1) % CMD_QUEUE_SIZE;
    if (cmd_queue_head == cmd_queue_tail) //buffer empty
      UCSR0B &= ~_BV(UDRIE0);             // disable TX interrupts
  }
*/
}

/*@}*/


