/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


#ifndef FUELCALC_H
#define FUELCALC_H

#include <inttypes.h>


struct corr_t {
  uint8_t ego;       // Oxygen Sensor Correction
  uint8_t air;       // Air Density Correction lookup
  uint8_t warm;      // Total Warmup Correction
  uint8_t baro;      // Barometer Lookup Correction
  uint16_t gammae_1k5;    // Total Gamma Enrichments, 1526 based
  uint16_t ve_hp;     // Current VE value & fraction from lookup table
  uint8_t lambda;    // lambda (afr)
  uint8_t tpsaccel;  // Acceleration enrichment
  uint8_t tpsfuelcut; // fuelcut
  uint8_t tps_overrun; // tps overrun fuelcut
  uint8_t als; // antilagsystem enrichment
};

#define GAMMAE_BASE (1526)

extern uint8_t tpsaclk;
extern struct search_table_t clt_idx;
extern struct corr_t corr;

void leave_cranking(void);
void fill_all_squirters(uint16_t pw);
void init_fuelcalc(void);
void search_clt_index(void);
void calc_parameters(void);
void baro_correction(uint8_t);
void warmup_enrich(void);
void tps_acc_enrich(void);
void o2(void);
void calc_total_enrichment(void);
uint8_t tps_overrun(void);
#ifdef ALS
#define als_deactivated (ALS_SWITCH_IN & _BV(ALS_SWITCH_PIN))
void als_enrich(void);
#endif // ALS

void calc_rpm(void);
void cranking(void);
void primepulse(void);


// overrun fuelcut bit flags
#define OVERRUN_ENABLE 0
#define OVERRUN_ACTIVE 1

#endif
