/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/

/*
  NOTE:
    The 'store to eeprom' functions has been totally rewritten.
    All the work are now done outside the interrupt routine
    as it should be.
*/

   /*!
   * \addtogroup g_storage EEPROM Memory Storage Routines
   */
/*@{*/

#include "storage.h"
#include "global.h"
#include "helpers.h"
#include "ve.h"
#include "ego.h"
#include "boostcontrol.h"

#include <inttypes.h>
#include <avr/eeprom.h>
#include <avr/signal.h>
#include <avr/interrupt.h>

//uint8_t WWURANGE[10] = {0, 20, 40, 60, 80, 100, 120, 140, 170, 200};
//uint8_t tpsdotrate[4] = {5, 20, 40, 77};

struct config_t config;
struct config_t config_ee SEEPROM;

struct ve_t ve;
struct ve_t ve_ee SEEPROM;

#ifndef NOIGN
struct ign_t ign;
struct ign_t ign_ee SEEPROM;
#endif

#ifdef HWMAP
struct hwmap_t hwmap;
struct hwmap_t hwmap_ee SEEPROM;
#endif

#ifdef BOOSTCONTROL
struct boost_t boost;
struct boost_t boost_ee SEEPROM;
#endif

/***************************************************************************/
/*** load config                                                         ***/
/***************************************************************************/
void load_variables(void) {

  /* read configuration variables from eeprom to sram */
  eeprom_read_block(&config, &config_ee, sizeof(struct config_t));
  eeprom_read_block(&ve, &ve_ee, sizeof(struct ve_t));
//  copyVEreftoVE();
  
#ifndef NOIGN
  eeprom_read_block(&ign, &ign_ee, sizeof(struct ign_t));
#endif

#ifdef HWMAP
  eeprom_read_block(&hwmap, &hwmap_ee, sizeof(struct hwmap_t));
#endif

#ifdef BOOSTCONTROL
  eeprom_read_block(&boost, &boost_ee, sizeof(struct boost_t));
#endif
}

/***************************************************************************/
/*** Read cell value from the EEPROM                                     ***/
/***************************************************************************/
void get_undovalue(uint8_t table, uint8_t idx) {
//TODO:check this why not working?...
	if (eeprom_flags & _BV(active))
		return; // EEPROM is active, don't bother it
	uint8_t *dest;
	uint8_t *source;
	
	switch(table)
	{
		case 'c':
			dest=(uint8_t*)&config;
			source=(uint8_t*)&config_ee;
			break;
		case 'j':
			// this is a special case because ve_table is a word array
			*(ve.ve_table + idx)=eeprom_read_word(ve_ee.ve_table + idx);
			return; // don't continue
		case 'k':
			dest=ve.kpa_range;
			source=ve_ee.kpa_range;
			break;
		case 'r':
			dest=ve.rpm_range;
			source=ve_ee.rpm_range;
			break;
		case 'l':
			dest=ve.lambda;
			source=ve_ee.lambda;		
			break;
		case 'd':
			dest=ve.ref_table;
			source=ve_ee.ref_table;
			break;
	#ifndef NOIGN
		case 'n':
			dest=ign.adv_table;
			source=ign_ee.adv_table;
			break;
	#endif
	#ifdef HWMAP
		case 'h':
			dest=hwmap.table;
			source=hwmap_ee.table;
			break;
	#endif
	#ifdef BOOSTCONTROL
		case 'b':
			dest=boost.boostmaxtarget_table;
			source=boost_ee.boostmaxtarget_table;
			break;
		case 't':
			dest=boost.tpstransfer_table;
			source=boost_ee.tpstransfer_table;
			break;
	#endif
		default:
			return; // do not change anything
	}
	
	*(dest + idx)=eeprom_read_byte((uint8_t *)(source + idx));	
}

/***************************************************************************/
/*** init storage                                                        ***/
/***************************************************************************/
void init_storage(void) {
  eeprom_flags = 0;
}


/***************************************************************************/
/*** save variables                                                      ***/
/***************************************************************************/
// call this function to store all variables
uint8_t save_variables(void) {
/* some stupid user corrections */
	
	if(config.ego_rich_limit > EGOCORR_BASE)
		config.ego_rich_limit = EGOCORR_BASE;	
		
	if(config.ego_lean_limit > EGOCORR_BASE)
		config.ego_lean_limit = EGOCORR_BASE;

  if (eeprom_flags & _BV(active))
    return -1; // busy storing configuration, take it easy!

  eeprom_flags |= _BV(start) | _BV(active);
  return 0;
}


/***************************************************************************/
/*** store control                                                       ***/
/***************************************************************************/
// place this in the mainloop, it polls the eeprom_flags to see if
// there is any work to do
void store_control(void) {
  if (eeprom_flags & _BV(active)) { // are the eeprom going to do anything at all
    if (eeprom_flags & _BV(start)) {   // any new work for me?
      eeprom_flags &= ~_BV(start);
      state = BEGIN_STORING;
#ifndef GENBOARDv3
      cli();
      ledstate |= _BV(LED_MISC); // turn on led
      sei();
#endif
    } else {
      if (!(eeprom_flags & _BV(busy))) { // has the write completed
	if (store_setup()) {
	  eeprom_flags &= ~_BV(active);  // done, clear flags
#ifndef GENBOARDv3
	  cli();
	  ledstate &= ~_BV(LED_MISC); // turn off led
	  sei();
#endif
	}
      }     
    }
  }
}


/***************************************************************************/
/*** store setup                                                         ***/
/***************************************************************************/
// setup pointers to sram/eeprom storage and call function 'store_bytes'
// returns 1 when done
uint8_t store_setup(void) {
  uint8_t ret_val=0;

  switch (state) {
  case BEGIN_STORING:
    // setup pointers for the config-structure
    p_eeprom = (uint8_t *)&config_ee;
    p_sram = (uint8_t *)&config;
    length = sizeof(struct config_t);
    offset = 0;
    state = STORE_CONFIG;
    break;

  case STORE_CONFIG:
    if (store_bytes()) {
      // setup pointers for the ve-structure
      p_eeprom = (uint8_t *)&ve_ee;
      p_sram = (uint8_t *)&ve;
      length = sizeof(struct ve_t);
      offset = 0;
      state = STORE_VE;
    }
    break;
	
  case STORE_VE:
#ifndef NOIGN
    if (store_bytes()) {
      // setup pointers for the ign-structure
      p_eeprom = (uint8_t *)&ign_ee;
      p_sram = (uint8_t *)&ign;
      length = sizeof(struct ign_t);
      offset = 0;
      state = STORE_IGN;
    }
    break;

  case STORE_IGN:
#endif
#ifdef HWMAP
    if (store_bytes()) {
      // setup pointers for the hwmap-structure
      p_eeprom = (uint8_t *)&hwmap_ee;
      p_sram = (uint8_t *)&hwmap;
      length = sizeof(struct hwmap_t);
      offset = 0;
      state = STORE_HWMAP;
    }
    break;

  case STORE_HWMAP:
#endif
#ifdef BOOSTCONTROL
    if (store_bytes()) {
      // setup pointers for the hwmap-structure
      p_eeprom = (uint8_t *)&boost_ee;
      p_sram = (uint8_t *)&boost;
      length = sizeof(struct boost_t);
      offset = 0;
      state = STORE_BOOST;
    }
    break;

  case STORE_BOOST:
#endif
    if (store_bytes()) {
      state = DONE;
    }
    break;

  default:
    ret_val = 1;
    break;
  }

  return ret_val;
}


/***************************************************************************/
/*** store bytes                                                         ***/
/***************************************************************************/
// returns 1 when entire stucture has been saved
uint8_t store_bytes(void) {
  uint8_t edata, sdata;

  edata = eeprom_read_byte(p_eeprom+offset);
  sdata = *(p_sram+offset);

  // search the structure until a difference is found between memory copy and eeprom,
  // or if the entire structure has been searched
  while ((edata == sdata) && (offset < (length-1))) {
    offset++;
    edata = eeprom_read_byte(p_eeprom+offset);
    sdata = *(p_sram+offset);
  }
  
  // store the value in eeprom 
  if (edata != sdata) { // test that the loop wasn't exited due to offset==sizeof...
    eeprom_write_byte(p_eeprom+offset, sdata);
    
    // set flags, and enable eeprom isr
    eeprom_flags |= _BV(busy);
    EECR |= _BV(EERIE);
  }
  
  if (offset >= (length-1)) {
    return 1;
  } else {
    offset++;
    return 0;
  }
}


/***************************************************************************/
/*** INTERRUPT: EEPROM ready                                             ***/
/***************************************************************************/
SIGNAL(SIG_EEPROM_READY) {
  eeprom_flags &= ~_BV(busy);  // signal completion of write
  EECR &= ~_BV(EERIE);         // disable interrupt
}

/*@}*/

