/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/

#ifndef __MEGASQUIRT_C__
#define __MEGASQUIRT_C__

#include "doxygen.h"
#include "vems.h"
#include "global.h"
#include "comm.h"
#include "fuelcalc.h"
#include "fuelcontrol.h"
#include "ringbuf.h"
#include "phasesignal.h"
#include "helpers.h"
#include "storage.h"
#include "menu.h"
#include "adc.h"
#include "actuators.h"
#include "tables.h"
#include "ve.h"
#include "iac.h"
#include "timing.h"
#include "ego.h"
#include "benchmark.h"
#include "wbo2.h"
#include "boostcontrol.h"
#include "sipr.h"

#ifdef  KEYBOARD
#include "keyboard.h"
#endif

#ifdef  PS2KEYBOARD
#include "kb.h"
#endif

#ifdef  LCD
#include "lcd.h"
#include "lcd_display.h"
#endif

#include <avr/io.h>
#include <inttypes.h>
#include <avr/wdt.h>
#include <stdlib.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/sfr_defs.h>
#include <avr/pgmspace.h>

#include "ioinit.c"

#ifdef SWITCHER
#include "switcher.c"
#endif

#include "spi.h"
#ifdef KNOCK_8101
#include "knock.h"
#endif

#ifdef CAPTURE_ALIEN_ECU
#include "ign_logging.h"
#endif

/*
TODO:
  Check reset source at power-up
  Watchdog timer 
*/

extern volatile struct engine_t engine;

uint8_t crank_timeout;
struct time_t rtc;

#include "eventqueue.h"
#include "lbus.c"
#include "edis.c"
#include "softpwm.c"
struct elapsed_t hundredms;
struct elapsed8_t sevenms;
struct elapsed8_t tenms;

#ifdef BOOSTCONTROL
struct elapsed8_t boostms;
#endif // BOOSTCONTROL
/* to avoid conflict with LCD pins, LEDPORT is always set from userspace,
 * not interrupt 
 */
volatile uint8_t ledstate;

/*!
 * \ingroup g_main_loop
 * \brief Set diagnostic LED's.
 *
 * \return none
 */
void setleds(void)
{
#ifdef GENBOARDv3
	// maybe via P259?
#else
#ifndef LED_MATRIX
	if(debug.stim_period != 0)	// we don't want to interfere with stim-out (see msp)
		return;

        LEDPORT = ((ledstate & LEDMASK 
#ifdef PINE2LED
       /* PE6 TACH signal is mirrored to LED_WARMUP to ease TACH-debug */
        & ~64) | (PINE & 64  
#endif
	)
        ) | (LEDPORT & ~LEDMASK);
#endif	// LED_MATRIX
#endif	// GENBOARDv3
}
/***************************************************************************/
/*** main entry point                                                    ***/
/***************************************************************************/
/*!
 * \ingroup g_main_loop
 * \brief Main entry point and cycle loop for fuel processing.
 *
 * \return none
 */
int main(void) {

  debug_init();

  // init HW and misc. variables
  init();

  // schedule a priming pulse
  primepulse();

  // init the SIPR library
  sipr_init();

  // synchronize timers
  counter_sync();
  // enable interrupts
  sei();

#ifdef PHASESIGNAL
	phase_idle();
#endif

// main loop
  for (;;) {
  
/* TODO: 
* introduce uint8_t scheduler_state (or mainloop_state?) variable.

In _every cycle_ do _all_ HPT that needs attention.
* high priority tasks (HPT):
** rpm_calc
** help the multitooth trigger code (currently all in interrupt)
** ... ?

And do max 1 LPT task that needs attention. Use the variable
to determine which is next.

* low priority tasks (LPT):
** all the rest (LCD, fuelcalc ... )

Note that currently several LPT could be executed before 
the code checks HPT-s.

Should we need 3 userspace priorities not 2 ?
(+ we have the highest realtime priority in eventqueue anyway)

*/
	calc_parameters();
	coolant_fan();
#ifdef GENBOARDv3
	water_pump();
#endif
	
	idle_solenoid();

#ifdef CAPTURE_ALIEN_ECU
	alien_logging_stimuli();
#endif

#ifdef WOT_OUTPUT_TPS_THRESHOLD
	if(debug.stim_period == 0)
		wot_output();
#endif

	rpm_output();

	setleds(); // make sure changes to ledstate will be reflected on the LEDs

    // only recalculate rpm if necessary
    if (engine.status_ext & _BV(new_rpm)) {
      cli();
      engine.status_ext &= ~_BV(new_rpm);
      sei();
      calc_rpm();
    }

    search_clt_index();
    search_rpm_kpa_index();

    // make sure the engine is spinning (what?? what's this condition for?)
    if (engine.rpm) {
      // only enter cranking mode on power-on or if rpm drops to zero. (why?)
      if ((engine.status_ext & _BV(crank_enable)) && (engine.rpm <= config.cranking_thres)) {
	cranking();
	//crank_timeout = CRANK_TIMEOUT; // do we want this here too?

	} else {    // running

		if(engine.status & _BV(crank))
			leave_cranking();

	// we need to be sure that 'crank_enable' isn't set due to some glitch in the rpm calculation
	// I hate this crank_timeout. Let's drop it.
	// With a glitch in the RPM calc, _everything_ goes bad anyway.
	if (crank_timeout == 1) {
	  crank_timeout=0;
	  cli();
	  engine.status_ext &= ~_BV(crank_enable);
	  sei();
	}

	// check if idle, and do idle control / anti-stall ?
    }
	
    } // engine.rpm
	
	// when do we want to reevaluate these?
	warmup_enrich();
	tps_acc_enrich();
#ifdef ALS	
	als_enrich();
#endif
	table_lookup();	// ve, ignadv, dwellcalc, etc...
	calc_total_enrichment();

#ifdef KNOCK_8101
	calc_knock_window();
	handle_knock();
#endif

	if (new_spark)
		calc_ign_advance();	// alien advance

#ifdef WBO2
    wbo2_heater();
#endif
	
    // every second
    if (rtc.flags & _BV(second)) {
		rtc.flags &= ~_BV(second);
		if (!(WBO2_SENSOR_CONFIGURED))
			ego_nbo2_ready();
		benchstats_maysend();

		if (engine.status & _BV(running)) {
		  if (crank_timeout > 1) // cranking timeout counter
		    crank_timeout--;
		}

    } // second

    // every 1/10th second    
    if (softelapsed(&hundredms, 25000)){
	    static uint8_t status_led;
		tpsaclk++;
		
		if (config.tpsdot_kpadot_conf & _BV(tps_or_kpa)) engine.last_tok= engine_kpa8();
		else engine.last_tok= engine.tps; // last tps or kpa for acc. enrichment
		
		// you want to switch off fuelpump if no rotation signal for (eg 2.56 sec)
		// Hopefully not because of an accident
		if(((etime.now_upper - (etime.ptrig_capturedtime_prev>>16)) > config.engine_off_delay)
			&& (etime.now_upper > config.pump_on_mintime)){ // but we don't want to switchoff at the very start
				cli();
				engine_off();
				sei();
		}
      // interrupt only touches rtc.flags, the rest is not protected
		if (--rtc.tsec == 0) {
			rtc.tsec = 10;
			rtc.sec++;
			rtc.flags |= _BV(second);	// it will be executed in next mainloop
		}
		
		if(config.iac_conf & _BV(precise_idle)){
			uint8_t dfi = debug.forced_iac;
			if(dfi){
				set_iac_pos(dfi);
			} else {
			// always calculate iac position (must always be ready to start the engine)
				calc_iac_pos();  
			}
		}
		
		if (engine.status & _BV(running)) {
			// Yeehaaaa, give me some VE values...
			if (config.ve_learn_conf & _BV(ve_learn_on))
				ve_learn();
		}

    // the status led indicates that the main loop hasn't stuck somewhere
	if (++status_led==4) {
		status_led=0;
		cli();
#ifdef GENBOARDv3	    
	    P259(DHIGH, P259_MISC);
#else	    
	    ledstate &= ~_BV(LED_STATUS);  //statusled on
#endif
	    sei();
    } else {
	    cli();
#ifdef GENBOARDv3	    
	    P259(DLOW, P259_MISC);
#else		    
	    ledstate |= _BV(LED_STATUS); //statusled off
#endif 
	    sei();
    }

      // process any timeouts in the receipt of a SIPR packet
      if (sipr_inc_timeout<255) {
        if (sipr_inc_timeout==0) {
          // reset the packet collector and send an NAK
          sipr_timeout();
        }
        else 
          sipr_inc_timeout--;      
      }

    }	// 100 ms
	
#ifdef BOOSTCONTROL
	{
	uint8_t i;
	// 20..30 Hz works best
	// 10-50hz range would be perfect.
	// we use period of 1..15 * (8 * 1024 usec)
	i = config.boost_conf;
	if (i && softelapsed8(&boostms, BOOSTCONTROL_PWM_PERIOD_1MS(i))){
		boostcontrol_act();
	}
	}
#endif // BOOSTCONTROL

    if (softelapsed8(&tenms, 15)){	// every 15 ms
#ifdef KNOCK_8101
	    decrease_knock_cnt(); // decrease time-based knock counter
#endif	
#ifdef LCD
if((debug.disable_flags & _BV(DIS_FLAG_LCD)) == 0){
    if (lcd.init) {
	    // initialize the lcd without a huge delay loop
		lcd_init_seq();
    } else {

	
		// write something to the cache
		lcd_do_something();
	}
	}
#endif
#ifdef MISC1OUT
	misc1_activity();
#endif //MISC1OUT
#ifdef MISC2OUT
	misc2_activity();
#endif //MISC2OUT
    } // 15 ms

    store_control(); // store variables to eeprom

    // rs232 buffer is always processed
    ringbuf_shift_proc(&menu_rs232, &comm_recbuf);

    // collect data for the transmit buffer
    // a lightweight scheduler could be included
    ringbuf_fill(&comm_sndbuf);


#ifdef  PS2KEYBOARD
	ps2proc_char();
#endif

#ifdef LCD
	if(lcd.init == 0) lcd_refresh();
#endif

// TODO: simple scheduler is on the corner (check the high priority stuff frequently)

#ifdef FIXIACFREQ
#define IACFREQ 6
#else
#define IACFREQ (1 + (config.iac_conf>>4))
#endif
	if (softelapsed8(&sevenms, IACFREQ)){
		if(config.iac_conf & _BV(stepper_iac)){
			move_iac();
		}

#ifdef WBO2
	if(!(WBO2_SENSOR_CONFIGURED))
#endif
		o2();

#ifndef NOIGN
		if (engine.ignstate == 4) { // previous IGNDEACT has enabled dwell
			uint16_t t;
			t = etime.now - engine.sparkat;
			if ((t < 0x8000) && (t > (engine.rpm_period<<1))) {

				// what the hell is going on?
				// checking for time spent in this ignstate == 4 
				// would be better than checking for very old trigger.

				// no trigger for 24 msec and we are in ignstate 4:
				// trigger was lost at high RPM.
				// let's not fry the coils, a spark is unavoidable
				cli();
				ign_deactivate_all();
				sei();
#ifdef BENCHMARK
				benchstats.ign_err++;
#endif
#ifdef V3_IGN_DRIVE_A
				test_ign_turnoff();
#endif
			}
		}
#endif
	} // if ... sevenms
	
#ifdef BENCHMARK
	benchstats.mainloopcnt++;
#endif
  } // main loop

  return (0);
} // main()


/***************************************************************************/
/*** init                                                                ***/
/***************************************************************************/
/*!
 * \ingroup g_init
 * \brief Main Initialization for power-up.
 * All under cli();
 * \return none
 */
void init(void) {
	ioinit();
#ifndef NOIGN
	ign_init();
#endif
#ifdef SWITCHER
	busy_square();
	ioinit();
#endif

#ifdef KEYBOARD
	keyboard_init();
#endif

#ifdef LCD
	lcd_init();
	//  update_chars("l1: -273C", 0);
	//  update_chars("l2: OK", 16); // changed 0 to O :-)
	lcd_display_mode(DISPLAY_NORMAL, &menu_rs232);
#endif

	ledstate=0;
  
	injector_pwm_init();

	eventqueue_init();

#ifdef IDLEPWM_HW
#ifdef GENBOARDv3
#error "conflict with injector pwm'ing"
#else 
#error "IDLEPWM_HW is not supported even on v2. Check config.iac_conf"
#endif
 
/* 
  Timer0 is set up to oparate in fast-pwm-mode. Prescaler is set to 256
  resulting pwm-frequency = 244.1 Hz (measured).

  Function to calculate pwm-frequency from prescaler and oscilator frequency is:

  F-PWM-Hz = F-OSC / (256 * OCRnMAXVAL)

  for this application this would result in:

  F-PWM-Hz = 16000000 / (256 * 256) = 244.141 Hz (Correct)
  
  This is necessary for the eventqueue anyway, so in eventqueue_init()
  Since COM0. bits also get set there, IDLEPWM_SOFT is not available (easy to change, but why?)
*/
  
  // Duty cylce is 40/256 (for testing only)
	OCR0 = 40;
  
  // turn on PortB idle_valve_pin 
    PORTB |= _BV(IDLEVALVE);

/*
  // setup timer0 to interrupt at 1 kHz (atmega128 datasheet page94) CTC mode
  // You can steel TMR0 for anything. Dave, do your IDLEPWM...
  TCCR0 = _BV(WGM01) | _BV(CS02); // CTC, div64 (page100)
  OCR0 = 249;
*/  
#endif 

  // setup external interrupt
    /*
#ifdef TACHBOTHEDGE 
	EICRB = _BV(ISC60);  // falling or rising edge
#else
	EICRB = _BV(ISC61);  // falling edge
#endif
    */

  // init rs232
  initUART();
	// menu system that is controllable from rs232 or ps2keyboard
	menu_init(&menu_rs232, MENU_ZERO);
	ringbuf_init(&comm_recbuf);
	ringbuf_init(&comm_sndbuf);
  
#ifdef COMM_AUTO_DUMP
  pushfuncpar(&srl_generic_P, 8, PSTR("HeloPingOK\r\n"));
#endif

  // load constants from eeprom to ram
  init_storage();
  load_variables();

#ifdef LCD
  lcd_display_setpage(config.lcd_default_view);
#endif

  // init the sensor array
  init_sensors();

  // get current barometric pressure
  set_sensors8(BARO, sensors8(MAP));
  baro_correction(sensors8(BARO));

  // init global variables
  calc_parameters();

	// are the zero assignments needed? gcc zeroes them, 
	// and we never call this other than at reset, do we?
	// BTW some parts are common with engine_off()
  engine.status = 0;
  engine.status_ext = _BV(o2_not_ready) | _BV(crank_enable);
  engine.status_trig = 0;
  engine.rpm = 0;
  engine.rpm_period = 0xFFFFFFFF;
  engine.rpm_vhr = 0;
  if (config.tpsdot_kpadot_conf & _BV(tps_or_kpa)) engine.last_tok= engine_kpa8();
  else engine.last_tok= engine.tps; // last tps or kpa for acc. enrichment

#ifndef NOIGN
	// TODO: remove, since useless. See table_lookup()
  engine.ignadv = config.ign_crank_advance;
	// should be called even without KNOCK_8101 defined... unless we rip adjust from everywhere
#ifdef KNOCK_8101
	reset_knock_adjust();
#endif	// KNOCK_8101
#endif	// NOIGN

  crank_timeout = CRANK_TIMEOUT;

#ifdef WBO2
  init_adc();  // adc is configured for free running mode etc. here
  init_wbo2();
#endif

  init_fuelcalc();

  init_ego();

  // init iac parameters
  init_iac();

  // enable fuelpump, 
  // it will be turned off within a couple of seconds, if engine isn't started
  fuelpump_on(); // we are still in initial cli(), good
  // done, we have pressure, the fuelpump is really fast!

	// beware: timing_init() must be called after init_wbo2()
	timing_init(); // should we call this earlier?  We want to catch early triggers but
	// we really would not like to use unitialized variables. Cannot do much, in cli() anyway...

#ifdef  PS2KEYBOARD
	// menu system that is controllable from ps2keyboard
	menu_init(&menu_ps2, MENU_mmm);
	init_kb();
#endif

#ifdef GENBOARDv3
	init_spi();
	init_spi_mcp3208();
#ifdef KNOCK_8101
	init_spi_knock();
	init_knock();
#endif
#endif

#ifdef CAPTURE_ALIEN_ECU
	init_alien_logging_stimuli();
#endif
	// initialize rtc variables:
//	rtc.tsec_clk = 10;
	rtc.tsec = 10;
//	rtc.iac_clk = 16;
	rtc.adc_clk = 1;

#ifdef BOOSTCONTROL
	boostcontrol_init();
#endif
} // init()

#endif // __MEGASQUIRT_C__


